PRO idl_example ; Tested on IDL v8.5.1

; Change these values to point to the product files
hdf5_file_path = "g3b.ssp.2021011831SSv05.30"
binary_file_path = "g3b.sspb.2021011831SSv05.30"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Plot an ozone profile (altitude vs. concentration) using
; a formatted (HDF5) product file
; Requires the HDF5 DLM (execute `HELP, 'hdf5', /DLM` for info)

; Open the HDF5 file
file_id = H5F_OPEN(hdf5_file_path)

; Isolate the dataset of interest
dataset_id = H5D_OPEN(file_id,"/Altitude Based Data/Aerosol Ozone Profiles/Ozone_AO3")
; HINT: execute `H5_LIST,hdf5_file_path` to list available datasets

; Read the data from the dataset
ozone_profile = H5D_READ(dataset_id)
; Read the fill value from the dataset attributes
ozone_fill_value = H5A_READ(H5A_OPEN_NAME(dataset_id,"_FillValue"))

; For plotting purposes, let's also grab altitudes and event ID using a simpler method
altitudes = H5_GETDATA(hdf5_file_path,"/Altitude Based Data/Altitude Information/Altitude")
event_id = H5_GETDATA(hdf5_file_path,"/Event Information/Event_ID")

; Establish a mask for fill values
mask = WHERE(ozone_profile LT ozone_fill_value)

; Plot the ozone profile
ozone_plot = PLOT(ozone_profile[mask],altitudes[mask],TITLE="Event "+event_id, $
   XTITLE="Ozone Concentration (cm^-3)", YTITLE="Altitude (km)", $
   XRANGE=[0,1e13])

; Save the plot
ozone_plot.SAVE, "ozone_"+event_id+".png"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Plot an aerosol profile (altitude vs. extinction coefficient) using
; an unformatted (binary) product file
; Requires read_solar.pro 

; Read the binary product file
read_solar,binary_file_path,status,data

; Aerosol extinction coefficient profiles are given for 9 channels, 
; lets use the data from the next-to-last (wavelength ~1 micron)
aerosol_profile = data.aerosolext[-2].aerext

; Read the fill value for float variables
aerosol_fill_value = data.flt_fill_value

; For plotting purposes, let's also grab the altitudes (aerosol only
; uses the first 90 of these), event ID, and aerosol wavelength
altitudes = data.altitude
event_id = data.event_id
wavelength = data.aer_wavelength[-2]

; Establish a mask for fill values
mask = WHERE(aerosol_profile LT aerosol_fill_value)

; Plot the aerosol extinction profile
aerosol_plot = PLOT(aerosol_profile[mask],altitudes[mask],TITLE="Event "+event_id, $
   XTITLE=string(wavelength,FORMAT="(F0.1)")+" nm Aerosol Extinction Coefficient (km^-1)", $
   YTITLE="Altitude (km)",/XLOG,XRANGE=[1e-6,1e-2])

aerosol_plot.SAVE, "aerosol_"+event_id+".png"

END
