"""
Notices
NumPy copyright statement and notice:
   Copyright (c) 2005-2018, NumPy Developers.
   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following
   conditions are met:
   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer. Redistributions in binary
   form must reproduce the above copyright notice, this list of conditions and
   the following disclaimer in the documentation and/or other materials
   provided with the distribution. Neither the name of the NumPy Developers nor
   the names of any contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.

h5py copyright statement and notice:
   Copyright (c) 2008-2013 Andrew Collette and contributors
   http://www.h5py.org
   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   a. Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
   b. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
   c. Neither the name of the author nor the names of contributors may be used
      to endorse or promote products derived from this software without specific
      prior written permission.
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT Not LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.

Python-HDF4 copyright statement and notice:
   Copyright (c) 2014 The Python-HDF4 Authors Permission is hereby granted, free
   of charge, to any person obtaining a copy of this software and associated
   documentation files (the "Software"), to deal in the Software without
   restriction, including without limitation the rights to use, copy, modify,
   merge, publish, distribute, sublicense, and/or sell copies of the Software,
   and to permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.


Disclaimers
No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF
PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
 
Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS
ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN
ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLESS THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR
ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.

Authors:     David Huber, SSAI <david.b.huber@nasa.gov>
             Carter Hulsey, SSAI <carter.b.hulsey@nasa.gov>
Date:        2/11/2016
             8/24/2018: Updated for v5.1 compatability an enabling of HDF5
             reading (Carter Hulsey)
Description: This module houses functions that read in SAGE III lunar L2 product
             files and return a data object containing the metadata and products
             for the given file type.  The functions are described below.
Notes:       This library has two dependencies:
             NumPy (mandatory) - Provides numerical functions and array creation
             SciPy (including NumPy) installation instructions can be found at
               https://scipy.org/install.html
             
             h5py(optional) - Library needed for reading hdf5 file formats
             Can be found at:
               https://github.com/h5py/h5py

             Python-HDF4 (optional) - A wrapper library for HDF4 functions.
             Installation instructions can be found at
               http://fhs.github.io/python-hdf4/install.html

             If you wish to work with HDF4/5 files, then you must have the
             HDF4/5 library installed as well as either Python-HDF4 or h5py.
             The prebuilt binaries for these libraries can be found at
               HDF4: https://portal.hdfgroup.org/display/support/HDF+4.2.14
               HDF5: https://www.hdfgroup.org/downloads/hdf5/

             Please report any bugs to david.b.huber@nasa.gov.

Examples:    Reading a level 2 HDF file
             from read_lunar import lun_l2_hdf5
             (data, stat) = lun_l2_hdf5("g3b.lsp.00612640v05.10")

             Reading a level 2 binary file
             from read_lunar import lun_l2_bin
             (data, stat) = lun_l2_bin("g3b.lspb.00612640v05.10")
"""

import numpy as np
from sys import version_info
import sys

# Declare which symbols to import
__version = version_info
if __version[0] == 2:
    __all__ = ["LunL2Data", "lun_l2_hdf4", "lun_l2_bin", "lun_l2_hdf5"]
elif __version[0] == 3:
    # Do not import the HDF4 readers for Python version 3.x.  A required module,
    # pyhdf, is not available for version 3 (at least at the time that this
    # module was written).
    __all__ = ["LunL2Data", "lun_l2_bin", "lun_l2_hdf5"]
else:
    print("Import error, unable to determine Python version")

# Set global constants
__c_MaxIndex = 20


class LunL2Data:
    """
    Class Name: LunL2Data
    Purpose:    The class contains the information from a given level 2 SAGE III
                data file.  This structure is the same for binary, HDF4, and HDF5
                files.
    """

    def __init__(self, vers, num_grnd_trk, num_bins, num_met_grid):
        # Structure for v5.1 and greater
        if vers >= 5.1:
            # Event information
            if vers >= 5.2:
                self.event_id = "XXXXXXXXXXXX"
                self.old_event_id = 0
            else:
                self.event_id = 0

            self.date = 0
            self.year_fraction = 0.0
            self.latitude = 0.0
            self.longitude = 0.0
            self.time = 0
            # Fill values
            self.int_fill_value = 0
            self.flt_fill_value = 0.0
            # Mission ID
            self.mission_id = 0
            # Software and data versions
            self.L0DO_Version = 0.0
            self.CCDTable_Version = 0
            self.L0_Version = 0.0
            self.Spectroscopic_DataBase_Version = 0.0
            self.GRAM95_Version = 0.0
            self.Met_Version = 0.0
            self.Software_Version = 0.0
            self.Dataproduct_Version = 0.0
            self.Lun_Model_Ver = 0.0
            self.Lun_Albedo_Ver = 0.0
            # Bin height
            self.bin_height = 0.0
            # Dimensions
            self.num_bins = num_bins
            self.num_met_grid = num_met_grid
            self.num_grnd_trk = num_grnd_trk
            # Event type
            self.sc_evt_type = 0
            self.gnd_evt_type = 0
            # Lunar beta angle
            self.BetaAngle_Lunar = 0.0
            # Observational information
            self.LunarPhase = 0.0
            self.ZenithAngle = 0.0
            self.aurora_flag = 0
            # Status flag
            self.Ephemeris_Source = 0
            # Ground track information
            self.gt_date = np.zeros((num_grnd_trk), dtype=np.int32)
            self.gt_time = np.zeros((num_grnd_trk), dtype=np.int32)
            self.gt_latitude = np.zeros((num_grnd_trk), dtype=np.float32)
            self.gt_longitude = np.zeros((num_grnd_trk), dtype=np.float32)
            self.gt_ray_dir = np.zeros((num_grnd_trk), dtype=np.float32)
            # Space Craft Information
            self.Space_Craft_Lat = np.zeros((num_grnd_trk), dtype=np.float32)
            self.Space_Craft_Lon = np.zeros((num_grnd_trk), dtype=np.float32)
            self.Space_Craft_Alt = np.zeros((num_grnd_trk), dtype=np.float32)
            # Altitude Information
            self.Altitude = np.zeros((num_bins), dtype=np.float32)
            self.geopotential_alt = np.zeros((num_bins), dtype=np.float32)
            # Temp/Press profile information
            self.Temperature = np.zeros((num_bins), dtype=np.float32)
            self.Temperature_uncert = np.zeros((num_bins), dtype=np.float32)
            self.Pressure = np.zeros((num_bins), dtype=np.float32)
            self.Pressure_uncert = np.zeros((num_bins), dtype=np.float32)
            self.Neutral_Density = np.zeros((num_bins), dtype=np.float32)
            self.Neutral_Density_uncert = np.zeros((num_bins), dtype=np.float32)
            self.Temp_Pressure_Source = np.zeros((num_bins), dtype=np.int32)
            # Tropopause temperature/altitude
            self.trop_temp = 0.0
            self.trop_alt = 0.0
            self.trop_press = 0.0
            # Meteorological inputs
            self.met_pressure = np.zeros((num_met_grid), dtype=np.float32)
            self.met_temp = np.zeros((num_met_grid), dtype=np.float32)
            self.met_temp_unc = np.zeros((num_met_grid), dtype=np.float32)
            self.met_altitude = np.zeros((num_met_grid), dtype=np.float32)
            self.met_source = 0
            # CCD and Bit Flag Information
            self.CCD_Temperature = 0.0
            self.Spectrometer_Zenith_Temperature = 0.0
            self.CCD_Temperature_minus_TEC = 0.0
            self.Ephemeris_Quality = 0
            self.SpecCalShift = 0.0
            self.SpecCalStretch = 0.0
            if vers >= 5.2:
                self.AzimuthAngle = np.zeros((2), dtype=np.float32)

            self.QAFlag = 0
            self.QAFlag_Altitude = np.zeros((num_bins), dtype=np.int32)
            # Decoded Bits
            self.HexErrFlag = 0
            self.ContWindowClosedFlag = 0
            self.TimeQualFlag = 0
            # self.DMPExoFlag = 0
            self.SpectralCalFlag = 0
            if vers >= 5.3:
                self.OffNadirFlag = 0
                self.TECFlag = 0
            # self.DMPAltFlag = np.zeros((num_bins), dtype = np.int32)
            # Altitude registration
            self.AbandAltRegQA = np.zeros((num_bins), dtype=np.int32)
            # A-Band altitude offset
            self.AbandAltOffset = 0.0
            # Composite derived ozone products
            self.Ozone = np.zeros((num_bins), dtype=np.float32)
            self.Ozone_uncert = np.zeros((num_bins), dtype=np.float32)
            self.Ozone_QA = np.zeros((num_bins), dtype=np.int32)
            # Nitrogen dioxide products
            self.NO2 = np.zeros((num_bins), dtype=np.float32)
            self.NO2_uncert = np.zeros((num_bins), dtype=np.float32)
            self.NO2_QA = np.zeros((num_bins), dtype=np.int32)
            # Nitrogen trioxide products
            self.NO3 = np.zeros((num_bins), dtype=np.float32)
            self.NO3_uncert = np.zeros((num_bins), dtype=np.float32)
            self.NO3_QA = np.zeros((num_bins), dtype=np.int32)
            # Chlorine dioxide products
            self.OClO = np.zeros((num_bins), dtype=np.float32)
            self.OClO_uncert = np.zeros((num_bins), dtype=np.float32)
            self.OClO_QA = np.zeros((num_bins), dtype=np.int32)
        # Structure of v5.0 and previous
        else:
            # Event information
            self.event_id = 0
            self.yr_day_tag = 0
            self.mission_time = 0
            # Fill values
            self.int_fill_value = 0
            self.flt_fill_value = 0.0
            # Mission ID
            self.mission_id = 0
            # Software and data versions
            self.L0DO_Version = 0.0
            self.L0_Version = 0.0
            self.DB_Version = 0.0
            self.GRAM95_Version = 0.0
            self.Met_Version = 0.0
            self.Software_Version = 0.0
            self.Dataproduct_Version = 0.0
            self.Lun_Model_Ver = 0.0
            self.Lun_Albedo_Ver = 0.0
            # Bin height
            self.bin_height = 0.0
            # Dimensions
            self.num_bins = num_bins
            self.num_met_grid = num_met_grid
            self.num_grnd_trk = num_grnd_trk
            # Event type
            self.sc_evt_type = 0
            self.gnd_evt_type = 0
            # Lunar beta angle
            self.BetaAngle = 0.0
            # Observational information
            self.LunarPhase = 0.0
            self.ZenithAngle = 0.0
            self.aurora_flag = 0
            # Status flag
            self.EvtStatusFlag = 0
            # Starting time/location
            self.st_date = 0
            self.st_time = 0
            self.st_latitude = 0.0
            self.st_longitude = 0.0
            self.st_altitude = 0.0
            # Ending time/location
            self.end_date = 0
            self.end_time = 0
            self.end_latitude = 0.0
            self.end_longitude = 0.0
            self.end_altitude = 0.0
            # Ground track information
            self.gt_date = np.zeros((num_grnd_trk), dtype=np.int32)
            self.gt_time = np.zeros((num_grnd_trk), dtype=np.int32)
            self.gt_latitude = np.zeros((num_grnd_trk), dtype=np.float32)
            self.gt_longitude = np.zeros((num_grnd_trk), dtype=np.float32)
            self.gt_ray_dir = np.zeros((num_grnd_trk), dtype=np.float32)
            # Tropopause temperature/altitude
            self.trop_temp = 0.0
            self.trop_alt = 0.0
            self.met_source = 0
            # Meteorological inputs
            self.met_pressure = np.zeros((num_met_grid), dtype=np.float32)
            self.met_temp = np.zeros((num_met_grid), dtype=np.float32)
            self.met_temp_unc = np.zeros((num_met_grid), dtype=np.float32)
            self.met_altitude = np.zeros((num_met_grid), dtype=np.float32)
            # Altitude registration
            self.geom_alt = np.zeros((num_bins), dtype=np.float32)
            self.AltRegQA = np.zeros((num_bins), dtype=np.int32)
            self.potential_alt = np.zeros((num_bins), dtype=np.float32)
            # A-Band altitude offset
            self.AbandAltOffset = 0.0
            # Retrieved T/P
            self.CurrentTemp = np.zeros((num_bins), dtype=np.float32)
            self.CurrentTemp_uncert = np.zeros((num_bins), dtype=np.float32)
            self.CurrentPress = np.zeros((num_bins), dtype=np.float32)
            self.CurrentPress_uncert = np.zeros((num_bins), dtype=np.float32)
            self.TPSource = np.zeros((num_bins), dtype=np.int32)
            # Composite derived ozone products
            self.CompositeO3 = np.zeros((num_bins), dtype=np.float32)
            self.CompositeO3_uncert = np.zeros((num_bins), dtype=np.float32)
            self.CompO3QA = np.zeros((num_bins), dtype=np.int32)
            # Nitrogen dioxide products
            self.NO2NDP = np.zeros((num_bins), dtype=np.float32)
            self.NO2NDP_uncert = np.zeros((num_bins), dtype=np.float32)
            self.NO2QA = np.zeros((num_bins), dtype=np.int32)
            # Nitrogen trioxide products
            self.NO3NDP = np.zeros((num_bins), dtype=np.float32)
            self.NO3NDP_uncert = np.zeros((num_bins), dtype=np.float32)
            self.NO3QA = np.zeros((num_bins), dtype=np.int32)
            # Chlorine dioxide products
            self.OClONDP = np.zeros((num_bins), dtype=np.float32)
            self.OClONDP_uncert = np.zeros((num_bins), dtype=np.float32)
            self.OClOQA = np.zeros((num_bins), dtype=np.int32)


##############################################################
def get_version(filename):
    """
    Name: get_version
    Author: Carter Hulsey <carter.b.hulsey@nasa.gov>
    Inputs: Filename
    Output: Version of the data product
    Description: Reads in filename and get version of data from
               filename. Returns Version as float.
    """
    splt_file = filename.split("v")
    vers = float(splt_file[-1])

    return vers


##############################################################


def __l_read_str(File, Count):
    """
    Name:        __l_read_str
    Inputs:      1) File handle or filename, 2) number of times to read
    Output:      The integer or integer array read in from the file
    Description: Reads in a string from a specified binary file on a
                 little-endian system.
    """

    return np.fromfile(File, dtype="S12", count=Count).newbyteorder()


##############################################################


def __l_read_int32(File, Count):
    """
    Name:        __l_read_int32
    Author:      David Huber <david.b.huber@nasa.gov>
    Inputs:      1) File handle or filename, 2) number of bytes to read in
                 3) number of times to read
    Output:      The integer or integer array read in from the file
    Description: Reads in a 4-byte integer from a specified binary file on a
                 little-endian system.
    """

    return np.fromfile(File, dtype=np.int32, count=Count).newbyteorder()


##############################################################


def __l_read_float32(File, Count):
    """
    Name:        __l_read_float32
    Author:      David Huber <david.b.huber@nasa.gov>
    Inputs:      1) File handle or filename, 2) size of 1-d array
    Output:      The integer or integer array read in from the file
    Description: Reads in a 4-byte floating point array from a specified binary
                 file on a little-endian system.
    """

    return np.fromfile(File, dtype=np.float32, count=Count).newbyteorder()


##############################################################


def __l_read_float64(File, Count):
    """
    Name:        __l_read_float64
    Author:      Darrell Dolliver
    Inputs:      1) File handle or filename, 2) size of 1-d array
    Output:      The integer or integer array read in from the file
    Description: Reads in a 8-byte floating point array from a specified binary
                 file on a little-endian system.
    """

    return np.fromfile(File, dtype=np.float64, count=Count).newbyteorder()


##############################################################


def __b_read_str(File, Count):
    """
    Name:        __b_read_str
    Inputs:      1) File handle or filename, 2) number of times to read
    Output:      The integer or integer array read in from the file
    Description: Reads in a string from a specified binary file on a
                 big-endian system.
    """

    return np.fromfile(File, dtype="S12", count=Count)


##############################################################


def __b_read_int32(File, Count):
    """
    Name:        __b_read_int32
    Author:      David Huber <david.b.huber@nasa.gov>
    Inputs:      1) File handle or filename, 2) number of bytes to read in
                 3) number of times to read
    Output:      The integer or integer array read in from the file
    Description: Reads in a 4-byte integer from a specified binary file on a
                 big-endian system.
    """

    return np.fromfile(File, dtype=np.int32, count=Count)


##############################################################


def __b_read_float32(File, Count):
    """
    Name:        __b_read_float32
    Author:      David Huber <david.b.huber@nasa.gov>
    Inputs:      1) File handle or filename, 2) size of 1-d array
    Output:      The integer or integer array read in from the file
    Description: Reads in a 4-byte floating point array from a specified binary
                 file on a big-endian system.
    """

    return np.fromfile(File, dtype=np.float32, count=Count)


##############################################################


def __b_read_float64(File, Count):
    """
    Name:        __b_read_float64
    Author:      Darrell Dolliver
    Inputs:      1) File handle or filename, 2) size of 1-d array
    Output:      The integer or integer array read in from the file
    Description: Reads in a 8-byte floating point array from a specified binary
                 file on a big-endian system.
    """

    return np.fromfile(File, dtype=np.float64, count=Count)


##############################################################


def lun_l2_hdf5(Filename):
    """
    Name:        lun_l2_hdf5
    Author:      David Huber <david.b.huber@nasa.gov>
    Inputs:      Filename of a lunar level 2 HDF5 product file
    Outputs:     1) A data object of type LunTranData containing the product data
                 2) A status flag indicating whether the read was successful or not
    Description: This function reads the data from a lunar HDF5 level 2 product
                 file and returns a data object containing the data from the file.
                 A status flag is also returned and should be checked before
                 attempting to access the structure after function completion.  A
                 status value of 0 indicates a successful read while a value of -1
                 indicates that an error occurred.
    """

    # Import h5py for interaction with HDF5 files
    import h5py

    # Get Version
    vers = get_version(Filename)

    # Only continue if version is less than 5.2
    if vers >= 5.2:
        print("Reader does not support version 5.2 HDF files")
        print("To use this reader for version 5.2 switch to binary file")
        print("Exiting...")
        sys.exit()

    # Determine if the input file exists
    try:
        File = h5py.File(Filename, "r")
    except OSError:
        print(Filename, "is not an HDF5 file. Exiting...\n")
        return (LunL2Data(1, 1, 1), -1)

    # Go ahead and declare the section names to avoid confusion later on when
    # reading in the file contents.
    Section1 = "Section 1.0 - File Header/Section 1.0 - File Header"
    Section2 = (
        "Section 2.0 - Product Identification/Section 2.0 - Product Identification"
    )
    Section3 = "Section 3.0 - File Description/File Description"
    Section4Root = "Section 4.0 - Event Identification"
    Section4 = Section4Root + "/Event Identification"
    Section41 = (
        Section4Root
        + "/Section 4.1 - Ground Track Data Over This Event"
        + "/Section 4.1 - Ground Track Data Over This Event"
    )
    Section42 = (
        Section4Root
        + "/Section 4.2 - Spacecraft Data Over This Event"
        + "/Section 4.2 - Spacecraft Data Over This Event"
    )
    Section43 = (
        Section4Root
        + "/Section 4.3 - Instrument Performance Data and Quality Flags For This Event"
        + "/Section 4.3 - Instrument Performance Data and Quality Flags For This Event"
    )
    Section44 = (
        Section4Root
        + "/Section 4.4 - Altitude Based BitFlags For This Event"
        + "/Section 4.4 - Altitude Based BitFlags For This Event"
    )
    Section5 = "Section 5.0 - Tropopause Data/Tropopause Data"
    Section51 = "Section 5.1 - Meteorological Data/Meteorological Data"
    Section6 = "Section 6.0 - Altitude-based array data/Altitude-based data"
    Section6A = (
        "Section 6.0A - Altitude-based scalar data/A-band Altitude Registration Offset"
    )
    Section61 = (
        "Section 6.1 - Temperature pressure profiles/Temperature pressure profiles"
    )
    Section62 = "Section 6.2 - Ozone profiles/Ozone profiles"
    Section63 = "Section 6.3 - Nitrogen Dioxide profiles/Nitrogen Dioxide profiles"
    Section64 = "Section 6.4 - Nitrogen Trioxide profiles/Nitrogen Trioxide profiles"
    Section65 = "Section 6.5 - OClO profiles/OClO profiles"

    # Make sure that we have the correct file type by checking whether all datasets exist
    e1 = Section1 in File
    e2 = Section2 in File
    e3 = Section3 in File
    e4 = Section4 in File
    e41 = Section41 in File
    e42 = Section42 in File
    e43 = Section43 in File
    e44 = Section44 in File
    e5 = Section5 in File
    e51 = Section51 in File
    e6 = Section6 in File
    e6a = Section6A in File
    e61 = Section61 in File
    e62 = Section62 in File
    e63 = Section63 in File
    e64 = Section64 in File
    e65 = Section65 in File

    if not (
        e1
        & e2
        & e3
        & e4
        & e41
        & e42
        & e43
        & e44
        & e5
        & e51
        & e6a
        & e61
        & e62
        & e63
        & e64
        & e65
    ):
        print(
            "Error reading "
            + str(Filename)
            + ": Unable to locate all SAGE datasets. Exiting...\n"
        )
        errlun = LunL2Data(1, 1, 1, 1)
        return (errlun, -1)

    # Get the file description handle
    DataFile = File[Section3][0]
    num_bins = DataFile[1]
    num_grnd_trk = DataFile[2]
    num_met_grid = DataFile[3]

    # Declare the output data object
    lunar = LunL2Data(vers, num_grnd_trk, num_bins, num_met_grid)

    # Get the height of each bin
    lunar.bin_height = DataFile[0]

    # Get the header information handle
    DataHeader = File[Section1][0]
    if vers <= 5.1:
        lunar.event_id = DataHeader[0]
        lunar.year_fraction = DataHeader[1]
        lunar.date = DataHeader[2]
        lunar.latitude = DataHeader[3]
        lunar.longitude = DataHeader[4]
        lunar.time = DataHeader[5]
        lunar.int_fill_value = DataHeader[6]
        lunar.flt_fill_value = DataHeader[7]
        lunar.mission_id = DataHeader[8]
    elif vers >= 5.2:
        lunar.event_id = DataHeader[0].decode("utf-8")
        lunar.old_event_id = DataHeader[1]
        lunar.year_fraction = DataHeader[2]
        lunar.date = DataHeader[3]
        lunar.latitude = DataHeader[4]
        lunar.longitude = DataHeader[5]
        lunar.time = DataHeader[6]
        lunar.int_fill_value = DataHeader[7]
        lunar.flt_fill_value = DataHeader[8]
        lunar.mission_id = DataHeader[9]

    # Get the product information handle
    DataProduct = File[Section2][0]
    lunar.L0DO_Version = DataProduct[0]
    lunar.CCDTable_Version = DataProduct[1]
    lunar.L0_Version = DataProduct[2]
    lunar.Spectroscopic_DataBase_Version = DataProduct[3]
    lunar.GRAM95_Version = DataProduct[4]
    lunar.Met_Version = DataProduct[5]
    lunar.Software_Version = DataProduct[6]
    lunar.Dataproduct_Version = DataProduct[7]
    lunar.Lun_Model_Ver = DataProduct[8]
    lunar.Lun_Albedo_Ver = DataProduct[9]

    # Get the event identification handle
    DataEvent = File[Section4][0]
    lunar.sc_evt_type = DataEvent[0]
    lunar.gnd_evt_type = DataEvent[1]
    lunar.BetaAngle_Lunar = DataEvent[2]
    lunar.LunarPhase = DataEvent[3]
    lunar.ZenithAngle = DataEvent[4]
    lunar.aurora_flag = DataEvent[5]
    lunar.Ephemeris_Source = DataEvent[6]

    # Get the event ground track information handle
    DataEventGround = np.asarray(File[Section41])
    lunar.gt_date = DataEventGround["Date (YYYYMMDD)"]
    lunar.gt_time = DataEventGround["Time (HHMMSS)"]
    lunar.gt_latitude = DataEventGround["Subtangent Latitude (0.0 +/- 90.0 deg)"]
    lunar.gt_longitude = DataEventGround["Subtangent Longitude (0.0 +/- 180.0 deg)"]
    lunar.gt_ray_dir = DataEventGround[
        "Ray Path Direction @ Subtan. Pt. (0.0 - 359.99 deg)"
    ]

    # Get the event Spacecraft  information handle
    DataEventSpacecraft = np.asarray(File[Section42])
    lunar.Space_Craft_Lat = DataEventSpacecraft[
        "Spacecraft Subtangent Latitude (0.0 +/- 90.0 deg)"
    ]
    lunar.Space_Craft_Lon = DataEventSpacecraft[
        "Spacecraft Subtangent Longitude (0.0 +/- 180.0 deg)"
    ]
    lunar.Space_Craft_Alt = DataEventSpacecraft[
        "Spacecraft Subtangent Altitude above Earth   (KM)"
    ]

    # Get Performance Data and Qual Flags
    DataQualFlag = File[Section43][0]
    lunar.CCD_Temperature = DataQualFlag[0]
    lunar.Spectrometer_Zenith_Temperature = DataQualFlag[1]
    lunar.CCD_Temperature_minus_TEC = DataQualFlag[2]
    lunar.Ephemeris_Quality = DataQualFlag[3]
    lunar.SpecCalShift = DataQualFlag[4]
    lunar.SpecCalStretch = DataQualFlag[5]
    if vers <= 5.1:
        lunar.QAFlag = DataQualFlag[6]
    else:
        lunar.AzimuthAngle[0] = DataQualFlag[6]
        lunar.AzimuthAngle[1] = DataQualFlag[7]
        lunar.QAFlag = DataQualFlag[8]

    # Get Alt Bases Bit Flags
    DataAltFlag = np.asarray(File[Section44])
    lunar.QAFlag_Altitude = DataAltFlag["Altitude Based Bit Flags"]

    # Get the tropopause data handle
    DataTrop = File[Section5][0]
    lunar.trop_alt = DataTrop[0]
    lunar.trop_temp = DataTrop[1]
    lunar.trop_press = DataTrop[2]
    lunar.met_source = DataTrop[3]

    # Get the meteorological input profiles handle
    DataMet = np.asarray(File[Section51])

    # Get the altitude registration product handle
    DataAlt = np.asarray(File[Section6])

    # Get the A-band altitude registration product handle
    DataABand = File[Section6A][0]
    lunar.AbandAltOffset = DataABand[0]

    # Get the temperature/pressure product handle
    DataTP = np.asarray(File[Section61])

    # Get the composite ozone product handle
    DataO3 = np.asarray(File[Section62])

    # Get the nitrogen dioxide product handle
    DataNO2 = np.asarray(File[Section63])

    # Get the nitrogen trioxide product handle
    DataNO3 = np.asarray(File[Section64])

    # Get the chlorine dioxide product handle
    DataOClO = np.asarray(File[Section65])

    # Populate the output data object
    # Section 5.1
    lunar.met_altitude = DataMet["Met Altitude (km)"]
    lunar.met_temp = DataMet["Met Temperature (K)"]
    lunar.met_temp_unc = DataMet["Met Temperature Unc (K)"]
    lunar.met_pressure = DataMet["Met Pressure (hPa)"]
    # Section 6
    lunar.Altitude = DataAlt["Geometric Altitude (km)"]
    lunar.AbandAltRegQA = DataAlt["Altitude Registration QA Flags"]
    lunar.geopotential_alt = DataAlt["Geopotential Altitude (km)"]

    # Section 6.1
    lunar.Temperature = DataTP["Current Temperature (K)"]
    lunar.Temperature_uncert = DataTP["Current Temperature Unc (K)"]
    lunar.Pressure = DataTP["Current Pressure (hPa)"]
    lunar.Pressure_uncert = DataTP["Current Pressure Unc (hPa)"]
    lunar.Neutral_Density = DataTP["Current Density (cm-3)"]
    lunar.Neutral_Density_uncert = DataTP["Current Density Unc (cm-3)"]
    lunar.Temp_Pressure_Source = DataTP["Pressure/Temp. Array Source"]

    # Section 6.2
    lunar.Ozone = DataO3["Ozone Concentration (cm-3)"]
    lunar.Ozone_uncert = DataO3["Ozone Concentration Unc (cm-3)"]
    lunar.Ozone_QA = DataO3["Ozone QA bit flags"]

    # Section 6.3
    lunar.NO2 = DataNO2["Nitrogen Dioxide Concentration (cm-3)"]
    lunar.NO2_uncert = DataNO2["Nitrogen Dioxide Concentration Unc (cm-3)"]
    lunar.NO2_QA = DataNO2["Nitrogen Dioxide QA bit flags"]

    # Section 6.4
    lunar.NO3 = DataNO3["Nitrogen Trioxide Concentration (cm-3)"]
    lunar.NO3_uncert = DataNO3["Nitrogen Trioxide Concentration Unc (cm-3)"]
    lunar.NO3_QA = DataNO3["Nitrogen Trioxide QA bit flags"]

    # Section 6.5
    lunar.OClO = DataOClO["OClO Concentration  (cm-3)"]
    lunar.OClO_uncert = DataOClO["OClO Concentration Unc (cm-3)"]
    lunar.OClO_QA = DataOClO["OClO QA bit flags"]

    # Close the HDF5 file
    File.close()

    # Decode Bits
    if (lunar.QAFlag & 2**0) != 0:
        lunar.HexErrFlag = 1
    if (lunar.QAFlag & 2**1) != 0:
        lunar.ContWindowClosedFlag = 1
    if (lunar.QAFlag & 2**2) != 0:
        lunar.TimeQualFlag = 1
    # if (lunar.QAFlag & 2**3) != 0:
    #    lunar.DMPExoFlag = 1
    if (lunar.QAFlag & 2**4) != 0:
        lunar.SpectralCalFlag = 1

    for i in range(0, lunar.num_bins):
        if (lunar.QAFlag_Altitude[i] & 2**0) != 0:
            lunar.DMPAltFlag[i] = 1

    return (lunar, 0)


########################################################################


# Read HDF4 level 2 product files
def lun_l2_hdf4(Filename):
    """
    Name:        lun_l2_hdf4
    Author:      David Huber <david.b.huber@nasa.gov>
    Inputs:      The filename of a lunar level 2 HDF4 product file
    Outputs:     1) A data object of type LunL2Data containing the product data
                 2) A status flag indicating whether the read was successful or not
    Description: This function reads the data from an HDF4 level 2 product file
                 and returns a data object containing the data from the file.
                 A status flag is also returned and should be checked before
                 attempting to access the structure after function completion.  A
                 status value of 0 indicates a success while a value of -1
                 indicates that an error occurred.
    """
    # Import pyhdf for interaction with HDF4 files.  Note that pyhdf is ONLY
    # available for Python 2.x.  If you wish to work with SAGE data files in
    # Python 3.x, use the binary or HDF5 readers.

    try:
        from pyhdf import HDF, VS
    except ImportError:
        if __version[0] == 3:
            raise ImportError(
                "Unable to locate the python-hdf4 library."
                + "  Please ensure version 0.9 or better is installed"
            )
        elif __version[0] == 2:
            raise ImportError(
                "Unable to locate the pyhdf or python-hdf4 library."
                + "  Please ensure version 0.9 (or higher) of either is installed."
            )

    # Open the file
    FileHandle = HDF.HDF(Filename)
    DataInterface = FileHandle.vstart()

    # Find the starting index
    for Index in range(__c_MaxIndex + 1):
        try:
            TestVD = DataInterface.attach(Index)
            TestVD.detach()
            break
        except HDF.HDF4Error:
            continue

    if Index == __c_MaxIndex:
        print("Failed to locate the starting VData index for ", Filename)
        return (-1, -1)

    # Get "Header" vdata information (Not actually the file header)
    VDataHeader = DataInterface.attach(Index)
    Index += 2
    try:
        HeaderData = VDataHeader.read()
    except HDF.HDF4Error:
        VDataHeader.detach()
        print("Failed to read header data from the HDF file")
        return (-1, -1)
    VDataHeader.detach()

    # Get product identification vdata information
    VDataProdId = DataInterface.attach(Index)
    Index += 2
    try:
        ProdIdData = VDataProdId.read()
    except HDF.HDF4Error:
        VDataProdId.detach()
        print("Failed to read product information from the HDF file")
        return (-1, -1)
    VDataProdId.detach()

    # Get file information
    VDataFileInfo = DataInterface.attach(Index)
    Index += 2
    try:
        ProdFileInfo = VDataFileInfo.read()
    except HDF.HDF4Error:
        VDataFileInfo.detach()
        print("Failed to read file information from the HDF file")
        return (-1, -1)
    VDataFileInfo.detach()

    # Get event information
    VDataEvent = DataInterface.attach(Index)
    Index += 2
    try:
        ProdEvent = VDataEvent.read()
    except HDF.HDF4Error:
        VDataEvent.detach()
        print("Failed to read event information from the HDF file")
        return (-1, -1)
    VDataEvent.detach()

    # Get information on the start of science data
    VDataSciStart = DataInterface.attach(Index)
    Index += 2
    try:
        ProdSciStart = VDataSciStart.read()
    except HDF.HDF4Error:
        VDataSciStart.detach()
        print("Failed to read science starting information from the HDF file")
        return (-1, -1)
    VDataSciStart.detach()

    # Get information on the end of the event
    VDataSciEnd = DataInterface.attach(Index)
    Index += 2
    try:
        ProdSciEnd = VDataSciEnd.read()
    except HDF.HDF4Error:
        VDataSciEnd.detach()
        print("Failed to read science ending information from the HDF file")
        return (-1, -1)
    VDataSciEnd.detach()

    # Collect array size information to initiate the output lunar object
    num_bins = ProdFileInfo[0][1]
    num_grnd_trk = ProdFileInfo[0][2]
    num_met_grid = ProdFileInfo[0][3]

    # Instatiate the lunar object
    lunar = LunL2Data(num_grnd_trk, num_bins, num_met_grid)

    # Begin populating the object attributes
    # Populate header data
    lunar.event_id = HeaderData[0][0]
    lunar.yr_day_tag = HeaderData[0][1]
    lunar.mission_time = HeaderData[0][2]
    lunar.int_fill_value = HeaderData[0][3]
    lunar.flt_fill_value = HeaderData[0][4]
    lunar.mission_id = HeaderData[0][5]
    # General product information
    lunar.L0DO_Version = ProdIdData[0][0]
    lunar.L0_Version = ProdIdData[0][1]
    lunar.DB_Version = ProdIdData[0][2]
    lunar.GRAM95_Version = ProdIdData[0][3]
    lunar.Met_Version = ProdIdData[0][4]
    lunar.Software_Version = ProdIdData[0][5]
    lunar.Dataproduct_Version = ProdIdData[0][6]
    lunar.Lun_Model_Ver = ProdIdData[0][7]
    lunar.Lun_Albedo_Ver = ProdIdData[0][8]
    # Array sizes
    lunar.bin_height = ProdFileInfo[0][0]
    # Event information
    lunar.sc_evt_type = ProdEvent[0][0]
    lunar.gnd_evt_type = ProdEvent[0][1]
    lunar.BetaAngle = ProdEvent[0][2]
    lunar.LunarPhase = ProdEvent[0][3]
    lunar.ZenithAngle = ProdEvent[0][4]
    lunar.aurora_flag = ProdEvent[0][5]
    lunar.EvtStatusFlag = ProdEvent[0][6]
    # Event starting information
    lunar.st_date = ProdSciStart[0][0]
    lunar.st_time = ProdSciStart[0][1]
    lunar.st_latitude = ProdSciStart[0][2]
    lunar.st_longitude = ProdSciStart[0][3]
    lunar.st_altitude = ProdSciStart[0][4]
    # Event ending information
    lunar.end_date = ProdSciEnd[0][0]
    lunar.end_time = ProdSciEnd[0][1]
    lunar.end_latitude = ProdSciEnd[0][2]
    lunar.end_longitude = ProdSciEnd[0][3]
    lunar.end_altitude = ProdSciEnd[0][4]

    # Get information on the ground track during the event
    VDataGround = DataInterface.attach(Index)
    Index += 2
    for trk in range(lunar.num_grnd_trk):
        try:
            ProdGround = VDataGround.read()
        except HDF.HDF4Error:
            VDataGround.detach()
            print("Failed to read ground track information from the HDF file")
            return (-1, -1)

        lunar.gt_date[trk] = ProdGround[0][0]
        lunar.gt_time[trk] = ProdGround[0][1]
        lunar.gt_latitude[trk] = ProdGround[0][2]
        lunar.gt_longitude[trk] = ProdGround[0][3]
        lunar.gt_ray_dir[trk] = ProdGround[0][4]

    VDataGround.detach()

    # Get the temperature/pressure profiles used during processing
    VDataTrop = DataInterface.attach(Index)
    Index += 2
    try:
        ProdTrop = VDataTrop.read()
    except HDF.HDF4Error:
        VDataTrop.detach()
        print("Failed to read T/P profiles used during processessing from the HDF file")
        return (-1, -1)

    lunar.trop_temp = ProdTrop[0][0]
    lunar.trop_alt = ProdTrop[0][1]
    lunar.met_source = ProdTrop[0][2]

    VDataTrop.detach()

    # Get input meteoroligcal data
    VDataMetData = DataInterface.attach(Index)
    Index += 2

    for alt in range(lunar.num_met_grid):
        try:
            ProdMetData = VDataMetData.read()
        except HDF.HDF4Error:
            VDataMetData.detach()
            print("Failed to read science metadata from the HDF file")
            return (-1, -1)

        lunar.met_pressure[alt] = ProdMetData[0][0]
        lunar.met_temp[alt] = ProdMetData[0][1]
        lunar.met_temp_unc[alt] = ProdMetData[0][2]
        lunar.met_altitude[alt] = ProdMetData[0][3]

    VDataMetData.detach()

    # Get science meta data
    VDataSciData = DataInterface.attach(Index)
    Index += 2

    for alt in range(lunar.num_bins):
        try:
            ProdSciData = VDataSciData.read()
        except HDF.HDF4Error:
            VDataSciData.detach()
            print("Failed to read science metadata from the HDF file")
            return (-1, -1)

        lunar.geom_alt[alt] = ProdSciData[0][0]
        lunar.AltRegQA[alt] = ProdSciData[0][1]
        lunar.potential_alt[alt] = ProdSciData[0][2]

    VDataSciData.detach()

    # Get the A-band altitude adjustment
    VDataAAdj = DataInterface.attach(Index)
    Index += 2
    try:
        ProdAAdj = VDataAAdj.read()
    except HDF.HDF4Error:
        VDataAAdj.detach()
        print("Failed to read T/P profiles used during processessing from the HDF file")
        return (-1, -1)

    lunar.AbandAltOffset = ProdAAdj[0][0]

    VDataAAdj.detach()

    # Get the temperature/pressure profiles used during processing
    VDataTempPres = DataInterface.attach(Index)
    Index += 2

    for alt in range(lunar.num_bins):
        try:
            ProdTempPres = VDataTempPres.read()
        except HDF.HDF4Error:
            VDataTempPres.detach()
            print(
                "Failed to read T/P profiles used during processessing from the HDF file"
            )
            return (-1, -1)

        lunar.CurrentTemp[alt] = ProdTempPres[0][0]
        lunar.CurrentTemp_uncert[alt] = ProdTempPres[0][1]
        lunar.CurrentPress[alt] = ProdTempPres[0][2]
        lunar.CurrentPress_uncert[alt] = ProdTempPres[0][3]
        lunar.TPSource[alt] = ProdTempPres[0][4]

    VDataTempPres.detach()

    # Get the composite ozone product
    VDataO3Comp = DataInterface.attach(Index)
    Index += 2
    for alt in range(lunar.num_bins):
        try:
            ProdO3Comp = VDataO3Comp.read()
        except HDF.HDF4Error:
            VDataO3Comp.detach()
            print("Failed to read the ozone profile from the HDF file")
            return (-1, -1)

        lunar.CompositeO3[alt] = ProdO3Comp[0][0]
        lunar.CompositeO3_uncert[alt] = ProdO3Comp[0][1]
        lunar.CompO3QA[alt] = ProdO3Comp[0][2]

    VDataO3Comp.detach()

    # Get the nitrogen dioxide product
    VDataNO2 = DataInterface.attach(Index)
    Index += 2
    for alt in range(lunar.num_bins):
        try:
            ProdNO2 = VDataNO2.read()
        except HDF.HDF4Error:
            VDataNO2.detach()
            print("Failed to read the ozone profile from the HDF file")
            return (-1, -1)

        lunar.NO2NDP[alt] = ProdNO2[0][0]
        lunar.NO2NDP_uncert[alt] = ProdNO2[0][1]
        lunar.NO2QA[alt] = ProdNO2[0][2]

    VDataNO2.detach()

    # Get the nitrogen trioxide product
    VDataNO3 = DataInterface.attach(Index)
    Index += 2
    for alt in range(lunar.num_bins):
        try:
            ProdNO3 = VDataNO3.read()
        except HDF.HDF4Error:
            VDataNO3.detach()
            print("Failed to read the Chappuis ozone product from the HDF file")
            return (-1, -1)

        lunar.NO3NDP[alt] = ProdNO3[0][0]
        lunar.NO3NDP_uncert[alt] = ProdNO3[0][1]
        lunar.NO3QA[alt] = ProdNO3[0][2]

    VDataNO3.detach()

    # Get the chlorine dioxide product
    VDataOClO = DataInterface.attach(Index)
    Index += 2
    for alt in range(lunar.num_bins):
        try:
            ProdOClO = VDataOClO.read()
        except HDF.HDF4Error:
            VDataOClO.detach()
            print("Failed to read the aerosol ozone product from the HDF file")
            return (-1, -1)

        lunar.OClONDP[alt] = ProdOClO[0][0]
        lunar.OClONDP_uncert[alt] = ProdOClO[0][1]
        lunar.OClOQA[alt] = ProdOClO[0][2]

    VDataOClO.detach()

    DataInterface.end()
    FileHandle.close()

    return (lunar, 0)


########################################################################


def lun_l2_bin(Filename):
    """
    Name:        lun_l2_bin
    Author:      David Huber <david.b.huber@nasa.gov>
    Inputs:      Filename of a lunar level 2 binary product file
    Outputs:     1) A data object of type LunL2Data containing the product data
                 2) A status flag indicating whether the read was successful or not
    Description: This function reads the data from a binary level 2 product file
                 and returns a data object containing the data from the file.
                 A status flag is also returned and should be checked before
                 attempting to access the structure after function completion.  A
                 status value of 0 indicates a success while a value of -1
                 indicates that an error occurred.
    """

    # Import byteorder from sys to determine system endianess
    from sys import byteorder

    EndianFlag = -1
    order = byteorder

    if order == "big":
        EndianFlag = 0
        read_int32 = __b_read_int32
        read_float32 = __b_read_float32
        read_float64 = __b_read_float64
        read_str = __b_read_str
    elif order == "little":
        EndianFlag = 1
        read_int32 = __l_read_int32
        read_float32 = __l_read_float32
        read_float64 = __l_read_float64
        read_str = __l_read_str
    else:
        print("Unable to determine endianness of this machine\n")
        return (-1, -1)

    # Get Version
    vers = get_version(Filename)

    # Open the file
    File = open(Filename, "r")

    if vers <= 5.0:
        # Retrieve scalar information from the product file
        event_id = read_int32(File, 1)[0]
        yr_day_tag = read_int32(File, 1)[0]
        mission_time = read_float32(File, 1)[0]
        int_fill_value = read_int32(File, 1)[0]
        flt_fill_value = read_float32(File, 1)[0]
        mission_id = read_int32(File, 1)[0]
        L0DO_Version = read_float32(File, 1)[0]
        L0_Version = read_float32(File, 1)[0]
        Software_Version = read_float32(File, 1)[0]
        Dataproduct_Version = read_float32(File, 1)[0]
        DB_Version = read_float32(File, 1)[0]
        GRAM95_Version = read_float32(File, 1)[0]
        Met_Version = read_float32(File, 1)[0]
        Lun_Model_Ver = read_float32(File, 1)[0]
        Lun_Albedo_Ver = read_float32(File, 1)[0]
        bin_height = read_float32(File, 1)[0]
        num_bins = read_int32(File, 1)[0]
        num_met_grid = read_int32(File, 1)[0]
        num_grnd_trk = read_int32(File, 1)[0]

        # Initialize the lunar object
        lunar = LunL2Data(vers, num_grnd_trk, num_bins, num_met_grid)

        # Fill the lunar object with the scalar data
        lunar.event_id = event_id
        lunar.yr_day_tag = yr_day_tag
        lunar.mission_time = mission_time
        lunar.int_fill_value = int_fill_value
        lunar.flt_fill_value = flt_fill_value
        lunar.mission_id = mission_id
        lunar.L0DO_Version = L0DO_Version
        lunar.L0_Version = L0_Version
        lunar.Software_Version = Software_Version
        lunar.Dataproduct_Version = Dataproduct_Version
        lunar.DB_Version = DB_Version
        lunar.GRAM95_Version = GRAM95_Version
        lunar.Met_Version = Met_Version
        lunar.Lun_Model_Ver = Lun_Model_Ver
        lunar.Lun_Albedo_Ver = Lun_Albedo_Ver
        lunar.bin_height = bin_height

        # Read the remainder of the data
        # Event type
        lunar.sc_evt_type = read_int32(File, 1)[0]
        lunar.gnd_evt_type = read_int32(File, 1)[0]
        # Beta angle
        lunar.BetaAngle = read_float32(File, 1)[0]
        # Lunar phase
        lunar.LunarPhase = read_float32(File, 1)[0]
        # Solar zenith angle
        lunar.ZenithAngle = read_float32(File, 1)[0]
        lunar.aurora_flag = read_int32(File, 1)[0]
        # Event status flag
        lunar.EvtStatusFlag = read_int32(File, 1)[0]
        # Starting time/position
        lunar.st_date = read_int32(File, 1)[0]
        lunar.st_time = read_int32(File, 1)[0]
        lunar.st_latitude = read_float32(File, 1)[0]
        lunar.st_longitude = read_float32(File, 1)[0]
        lunar.st_altitude = read_float32(File, 1)[0]
        # Ending time/position
        lunar.end_date = read_int32(File, 1)[0]
        lunar.end_time = read_int32(File, 1)[0]
        lunar.end_latitude = read_float32(File, 1)[0]
        lunar.end_longitude = read_float32(File, 1)[0]
        lunar.end_altitude = read_float32(File, 1)[0]
        # Ground track information
        lunar.gt_date = read_int32(File, lunar.num_grnd_trk)
        lunar.gt_time = read_int32(File, lunar.num_grnd_trk)
        lunar.gt_latitude = read_float32(File, lunar.num_grnd_trk)
        lunar.gt_longitude = read_float32(File, lunar.num_grnd_trk)
        lunar.gt_ray_dir = read_float32(File, lunar.num_grnd_trk)
        # Tropopause temperature/altitude
        lunar.trop_temp = read_float32(File, 1)[0]
        lunar.trop_alt = read_float32(File, 1)[0]
        # Meteorological inputs
        lunar.met_pressure = read_float32(File, lunar.num_met_grid)
        lunar.met_temp = read_float32(File, lunar.num_met_grid)
        lunar.met_temp_unc = read_float32(File, lunar.num_met_grid)
        lunar.met_altitude = read_float32(File, lunar.num_met_grid)
        lunar.met_source = read_int32(File, 1)[0]
        # Altitude registration
        lunar.geom_alt = read_float32(File, lunar.num_bins)
        lunar.AltRegQA = read_int32(File, lunar.num_bins)
        lunar.potential_alt = read_float32(File, lunar.num_bins)
        # A-band altitude adjustment
        lunar.AbandAltOffset = read_float32(File, 1)[0]
        # Temperature/pressure profiles
        lunar.CurrentTemp = read_float32(File, lunar.num_bins)
        lunar.CurrentTemp_uncert = read_float32(File, lunar.num_bins)
        lunar.CurrentPress = read_float32(File, lunar.num_bins)
        lunar.CurrentPress_uncert = read_float32(File, lunar.num_bins)
        lunar.TPSource = read_int32(File, lunar.num_bins)
        # Composite O3 products
        lunar.CompositeO3 = read_float32(File, lunar.num_bins)
        lunar.CompositeO3_uncert = read_float32(File, lunar.num_bins)
        lunar.CompO3QA = read_int32(File, lunar.num_bins)
        # Nitrogen Dioxide products
        lunar.NO2NDP = read_float32(File, lunar.num_bins)
        lunar.NO2NDP_uncert = read_float32(File, lunar.num_bins)
        lunar.NO2QA = read_int32(File, lunar.num_bins)
        # Nitrogen Trioxide products
        lunar.NO3NDP = read_float32(File, lunar.num_bins)
        lunar.NO3NDP_uncert = read_float32(File, num_bins)
        lunar.NO3QA = read_int32(File, lunar.num_bins)
        # Chlorine Dioxide products
        lunar.OClONDP = read_float32(File, lunar.num_bins)
        lunar.OClONDP_uncert = read_float32(File, lunar.num_bins)
        lunar.OClOQA = read_int32(File, lunar.num_bins)

        File.close()
        return (lunar, 0)
    else:
        # Retrieve scalar information from the product file
        if vers >= 5.2:
            event_id = read_str(File, 1)[0]
            old_event_id = read_int32(File, 1)[0]
        else:
            event_id = read_int32(File, 1)[0]

        date = read_int32(File, 1)[0]
        if vers >= 5.3:
            fraction_time = read_float64(File, 1)[0]
        else:
            fraction_time = read_float32(File, 1)[0]
        latitude = read_float32(File, 1)[0]
        longitude = read_float32(File, 1)[0]
        time = read_int32(File, 1)[0]
        int_fill_value = read_int32(File, 1)[0]
        flt_fill_value = read_float32(File, 1)[0]
        mission_id = read_int32(File, 1)[0]
        L0DO_Version = read_float32(File, 1)[0]
        CCDVersion = read_int32(File, 1)[0]
        L0_Version = read_float32(File, 1)[0]
        Software_Version = read_float32(File, 1)[0]
        Dataproduct_Version = read_float32(File, 1)[0]
        Spectroscopic_DataBase_Version = read_float32(File, 1)[0]
        GRAM95_Version = read_float32(File, 1)[0]
        Met_Version = read_float32(File, 1)[0]
        Lun_Model_Ver = read_float32(File, 1)[0]
        Lun_Albedo_Ver = read_float32(File, 1)[0]
        bin_height = read_float32(File, 1)[0]
        num_bins = read_int32(File, 1)[0]
        num_met_grid = read_int32(File, 1)[0]
        num_grnd_trk = read_int32(File, 1)[0]
        # Initialize the lunar object
        lunar = LunL2Data(vers, num_grnd_trk, num_bins, num_met_grid)

        # Fill the lunar object with the scalar data
        if vers >= 5.2:
            lunar.event_id = event_id.decode("utf-8")
            lunar.old_event_id = old_event_id
        else:
            lunar.event_id = event_id

        lunar.date = date
        lunar.year_fraction = fraction_time
        lunar.latitude = latitude
        lunar.longitude = longitude
        lunar.time = time
        lunar.int_fill_value = int_fill_value
        lunar.flt_fill_value = flt_fill_value
        lunar.mission_id = mission_id
        lunar.L0DO_Version = L0DO_Version
        lunar.CCDTable_Version = CCDVersion
        lunar.L0_Version = L0_Version
        lunar.Software_Version = Software_Version
        lunar.Dataproduct_Version = Dataproduct_Version
        lunar.Spectroscopic_DataBase_Version = Spectroscopic_DataBase_Version
        lunar.GRAM95_Version = GRAM95_Version
        lunar.Met_Version = Met_Version
        lunar.Lun_Model_Ver = Lun_Model_Ver
        lunar.Lun_Albedo_Ver = Lun_Albedo_Ver
        lunar.bin_height = bin_height

        # Read the remainder of the data
        # Event type
        lunar.sc_evt_type = read_int32(File, 1)[0]
        lunar.gnd_evt_type = read_int32(File, 1)[0]
        # Beta angle
        lunar.BetaAngle_Lunar = read_float32(File, 1)[0]
        # Lunar phase
        lunar.LunarPhase = read_float32(File, 1)[0]
        # Solar zenith angle
        lunar.ZenithAngle = read_float32(File, 1)[0]
        lunar.aurora_flag = read_int32(File, 1)[0]
        # Event status flag
        lunar.Ephemeris_Source = read_int32(File, 1)[0]
        # Ground track information
        lunar.gt_date = read_int32(File, lunar.num_grnd_trk)
        lunar.gt_time = read_int32(File, lunar.num_grnd_trk)
        lunar.gt_latitude = read_float32(File, lunar.num_grnd_trk)
        lunar.gt_longitude = read_float32(File, lunar.num_grnd_trk)
        lunar.gt_ray_dir = read_float32(File, lunar.num_grnd_trk)
        # Space Craft Information
        lunar.Space_Craft_Lat = read_float32(File, lunar.num_grnd_trk)
        lunar.Space_Craft_Lon = read_float32(File, lunar.num_grnd_trk)
        lunar.Space_Craft_Alt = read_float32(File, lunar.num_grnd_trk)
        # Altitude Information
        lunar.Altitude = read_float32(File, lunar.num_bins)
        lunar.geopotential_alt = read_float32(File, lunar.num_bins)
        # Temp/Press profile Information
        lunar.Temperature = read_float32(File, lunar.num_bins)
        lunar.Temperature_uncert = read_float32(File, lunar.num_bins)
        lunar.Pressure = read_float32(File, lunar.num_bins)
        lunar.Pressure_uncert = read_float32(File, lunar.num_bins)
        lunar.Neutral_Density = read_float32(File, lunar.num_bins)
        lunar.Neutral_Density_uncert = read_float32(File, lunar.num_bins)
        lunar.Temp_Pressure_Source = read_int32(File, lunar.num_bins)
        # Tropopause temperature/altitude
        lunar.trop_temp = read_float32(File, 1)[0]
        lunar.trop_alt = read_float32(File, 1)[0]
        lunar.trop_press = read_float32(File, 1)[0]
        # Meteorological inputs
        lunar.met_pressure = read_float32(File, lunar.num_met_grid)
        lunar.met_temp = read_float32(File, lunar.num_met_grid)
        lunar.met_temp_unc = read_float32(File, lunar.num_met_grid)
        lunar.met_altitude = read_float32(File, lunar.num_met_grid)
        lunar.met_source = read_int32(File, 1)[0]
        # CCD and bit flag Info
        lunar.CCD_Temperature = read_float32(File, 1)[0]
        lunar.Spectrometer_Zenith_Temperature = read_float32(File, 1)[0]
        lunar.CCD_Temperature_minus_TEC = read_float32(File, 1)[0]
        lunar.Ephemeris_Quality = read_int32(File, 1)[0]
        lunar.SpecCalShift = read_float32(File, 1)[0]
        lunar.SpecCalStretch = read_float32(File, 1)[0]
        if vers >= 5.2:
            lunar.AzimuthAngle = read_float32(File, 2)

        lunar.QAFlag = read_int32(File, 1)[0]
        lunar.QAFlag_Altitude = read_int32(File, lunar.num_bins)
        # Altitude registration
        lunar.AbandAltRegQA = read_int32(File, lunar.num_bins)
        # A-band altitude adjustment
        lunar.AbandAltOffset = read_float32(File, 1)[0]
        # Composite O3 products
        lunar.Ozone = read_float32(File, lunar.num_bins)
        lunar.Ozone_uncert = read_float32(File, lunar.num_bins)
        lunar.Ozone_QA = read_int32(File, lunar.num_bins)
        # Nitrogen Dioxide products
        lunar.NO2 = read_float32(File, lunar.num_bins)
        lunar.NO2_uncert = read_float32(File, lunar.num_bins)
        lunar.NO2_QA = read_int32(File, lunar.num_bins)
        # Nitrogen Trioxide products
        lunar.NO3 = read_float32(File, lunar.num_bins)
        lunar.NO3_uncert = read_float32(File, num_bins)
        lunar.NO3_QA = read_int32(File, lunar.num_bins)
        # Chlorine Dioxide products
        lunar.OClO = read_float32(File, lunar.num_bins)
        lunar.OClO_uncert = read_float32(File, lunar.num_bins)
        lunar.OClO_QA = read_int32(File, lunar.num_bins)

        File.close()

        # Decode Bits
        if (lunar.QAFlag & 2**0) != 0:
            lunar.HexErrFlag = 1
        if (lunar.QAFlag & 2**1) != 0:
            lunar.ContWindowClosedFlag = 1
        if (lunar.QAFlag & 2**2) != 0:
            lunar.TimeQualFlag = 1
        # if (lunar.QAFlag & 2**3) != 0:
        #  lunar.DMPExoFlag = 1
        if (lunar.QAFlag & 2**4) != 0:
            lunar.SpectralCalFlag = 1
        if vers >= 5.3:
            if (lunar.QAFlag & 2**5) != 0:
                lunar.OffNadirFlag = 1
            if (lunar.QAFlag & 2**6) != 0:
                lunar.TECFlag = 1

        # for i in range(0,lunar.num_bins):
        #  if (lunar.QAFlag_Altitude[i] & 2**0) != 0:
        # lunar.DMPAltFlag[i] = 1

        return (lunar, 0)
